/* JFlex example: part of Java language lexer specification */
import java_cup.runtime.*;

/**
 * This class is a simple example lexer.
 */
%%
%class Lexer
%unicode
%cup
%debug
%line
%column
%{
  StringBuffer string = new StringBuffer();

  private Symbol symbol(int type) {
    return new Symbol(type, yyline, yycolumn);
  }
  private Symbol symbol(int type, Object value) {
    return new Symbol(type, yyline, yycolumn, value);
  }
%}
LineTerminator = \r|\n|\r\n
InputCharacter = [^\r\n]
WhiteSpace     = {LineTerminator} | [ \t\f]

/* comments */
Comment = {TraditionalComment} | {EndOfLineComment} | {DocumentationComment}

TraditionalComment   = "/*" [^*] ~"*/" | "/*" "*"+ "/"
EndOfLineComment     = "//" {InputCharacter}* {LineTerminator}
DocumentationComment = "/**" {CommentContent} "*"+ "/"
CommentContent       = ( [^*] | \*+ [^/*] )*

Identifier = [:jletter:] [:jletterdigit:]*

DecIntegerLiteral = 0 | [1-9][0-9]*
%state STRING

%%
/* keywords */
<YYINITIAL> "struct"             { return symbol(sym.STRUCT); }
<YYINITIAL> "const"              { return symbol(sym.CONST); }
<YYINITIAL> "else"               { return symbol(sym.ELSE); }
<YYINITIAL> "while"              { return symbol(sym.WHILE); }
<YYINITIAL> "if"                 { return symbol(sym.IF); }
<YYINITIAL> "abstract"           { return symbol(sym.ABSTRACT); }
<YYINITIAL> "boolean"            { return symbol(sym.BOOLEAN); }
<YYINITIAL> "break"              { return symbol(sym.BREAK); }
<YYINITIAL> "long"               { return symbol(sym.SYM_LONG); }
<YYINITIAL> "int"                { return symbol(sym.SYM_INT); }

<YYINITIAL> "void"               { return symbol(sym.VOID); }
<YYINITIAL> {
  /* identifiers */ 
  {Identifier}                   { return symbol(sym.IDENTIFIER, new String(yytext())); }
 
  /* literals */
  {DecIntegerLiteral}            { return symbol(sym.INTEGER_LITERAL); }
  \"                             { string.setLength(0); yybegin(STRING); }

  /* operators */
  "!="                           { return symbol(sym.NEQ); }
  "="                            { return symbol(sym.EQ); }
  "=="                           { return symbol(sym.EQEQ); }
  "+"                            { return symbol(sym.PLUS); }
  "-"                            { return symbol(sym.MINUS); }
  "*"                            { return symbol(sym.ASTERICK); }
  "%"                            { return symbol(sym.MOD); }
  "/"                            { return symbol(sym.FSLASH); }
  "\\"                           { return symbol(sym.BSLASH); }
  "<"                            { return symbol(sym.LESS); }
  "<="                           { return symbol(sym.LESSEQ); }
  ">"                            { return symbol(sym.GREATER); }
  ">="                           { return symbol(sym.GREATEREQ); }
  "("                            { return symbol(sym.LPAREN); }
  ")"                            { return symbol(sym.RPAREN); }
  "{"                            { return symbol(sym.LBRACE); }
  "}"                            { return symbol(sym.RBRACE); }
  "["                            { return symbol(sym.LSQPAREN); }
  "]"                            { return symbol(sym.RSQPAREN); }
  ","                            { return symbol(sym.COMMA); }
  ";"                            { return symbol(sym.SEMICOLON); }
 

  /* comments */
  {Comment}                      { /* ignore */ }
 
  /* whitespace */
  {WhiteSpace}                   { /* ignore */ }
}
<STRING> {
  \"                             { yybegin(YYINITIAL); 
                                   return symbol(sym.STRING_LITERAL, 
                                   string.toString()); }
  [^\n\r\"\\]+                   { string.append( yytext() ); }
  \\t                            { string.append('\t'); }
  \\n                            { string.append('\n'); }

  \\r                            { string.append('\r'); }
  \\\"                           { string.append('\"'); }
  \\                             { string.append('\\'); }
}
/* error fallback */
.|\n                             { throw new Error("Illegal character <"+
                                                    yytext()+">"); }
