package compiler.ast;

import java.util.HashMap;
import java.util.Map;
import java.util.LinkedList;

/**
 * Symbol table.
 */
public class SymbolTable {
  public enum VarKind { LOCAL, GLOBAL }

  private Map<Symbol,String> labels;
  private Map<Symbol,Offset> offsets;
  private Map<Symbol,Type> types;

  private LinkedList<Map<String,Symbol>> context;
  private LinkedList<Symbol> currentMethod;
  private Map<String,Symbol> methods;  // only one global method scope

  private Map<Symbol,Integer> formals, locals;
  private int numFields;


  public class Offset {
    private VarKind kind;
    private int offset;
    public Offset(VarKind kind, int offset) {
      this.kind = kind;
      this.offset = offset;
    }
    public VarKind getKind() { return kind; }
    public int getOffset() { return offset; }
  }

  public SymbolTable() {
    formals = new HashMap<Symbol,Integer>();
    locals = new HashMap<Symbol,Integer>();
    
    context = new LinkedList<Map<String,Symbol>>();
    currentMethod = new LinkedList<Symbol>();
    methods = new HashMap<String,Symbol>();

    offsets = new HashMap<Symbol,Offset>();
    labels = new HashMap<Symbol,String>();
    types = new HashMap<Symbol,Type>();

    // Root context
    context.add(new HashMap<String,Symbol>());
    currentMethod.add(null);
  }

  public String getMethodLabel(Symbol name) { return labels.get(name); }
  public Offset getVarOffset(Symbol name) { return offsets.get(name); }
  public Type getType(Symbol name) { return types.get(name); }

  private Symbol makeSymbol(String key) {
    Map<String,Symbol> current = context.getFirst();
    if (current.containsKey(key))
      return null;
    Symbol s = new Symbol();
    current.put(key, s);
    return s;
  }

  private Symbol getSymbol(String key) {
    for (Map<String,Symbol> map : context) {
      if (map.containsKey(key))
        return map.get(key);
    }
    return null;
  }

  public void declareFormal(Type t, String id) {
    Symbol s = makeSymbol(id);
    if (s == null)
      throw new Error(String.format("Duplicate formal name: %s", id));

    incrementFormals(getCurrentMethod());
    int off = -formals.get(getCurrentMethod());
    offsets.put(s, new Offset(VarKind.LOCAL, off));
    types.put(s, t);
  }

  public void declareLocal(Type t, String id) {
    Symbol s = makeSymbol(id);
    if (s == null)
      throw new Error(String.format("Duplicate local name: %s", id));

    incrementLocals(getCurrentMethod());
    int off = locals.get(getCurrentMethod()) + 1; // +1 for saved PC
    offsets.put(s, new Offset(VarKind.LOCAL, off));
    types.put(s, t);
  }

  public void declareField(Type t, String id) {
    Symbol s = makeSymbol(id);
    if (s == null)
      throw new Error(String.format("Duplicate variable name: %s", id));

    numFields++;
    offsets.put(s, new Offset(VarKind.GLOBAL, numFields));
    types.put(s, t);
  }

  public void declareMethod(Type t, String id) {
    if (methods.containsKey(id))
      throw new Error(String.format("Duplicate method name: %s", id));
    
    Symbol s = new Symbol();
    methods.put(id, s);
    labels.put(s, String.format("L%d", methods.size()));
    types.put(s, t);
  }

  public Symbol getMethodSymbol(String id) {
    Symbol s = methods.get(id);
    if (s == null)
      throw new Error(String.format("Unknown method name: %s", id));

    return s;
  }

  public Symbol getVariableSymbol(String id) {
    Symbol s = getSymbol(id);
    if (s == null)
      throw new Error(String.format("Unknown variable name: %s", id));
    
    return s;
  }

  public int getNumLocals(Symbol method) {
    return locals.containsKey(method) ? locals.get(method) : 0;
  }

  public int getNumFormals(Symbol method) {
    return formals.containsKey(method) ? formals.get(method) : 0;
  }

  public int getNumFields() { return numFields; }

  public void pushMethodContext(Symbol method) {
    currentMethod.addFirst(method);
    context.addFirst(new HashMap<String,Symbol>());
  }

  public void pushVariableContext() {
    currentMethod.addFirst(getCurrentMethod());
    context.addFirst(new HashMap<String,Symbol>());
  }

  public void popContext() {
    currentMethod.removeFirst();
    context.removeFirst();
  }

  private Symbol getCurrentMethod() { return currentMethod.getFirst(); }
  private void incrementLocals(Symbol m) { incrementMap(locals, m); }
  private void incrementFormals(Symbol m) { incrementMap(formals, m); }
  private void incrementMap(Map<Symbol,Integer> map, Symbol method) {
    int value = map.containsKey(method) ? map.get(method) : 0;
    map.put(method, value + 1);
  }
}
